from gym.envs.registration import register

register(
    id="mobile-base-vel-v0",
    entry_point="planarenvs.mobile_base.envs:MobileBaseVelEnv",
)
register(
    id="mobile-base-acc-v0",
    entry_point="planarenvs.mobile_base.envs:MobileBaseAccEnv",
)



from MotionPlanningEnv.sphereObstacle import SphereObstacle
from MotionPlanningEnv.dynamicSphereObstacle import DynamicSphereObstacle

obst1Dict = {
    "type": "sphere",
    "geometry": {"position": [0.0, 5.0], "radius": 1.0},
}
sphereObst1 = SphereObstacle(name="simpleSphere", content_dict=obst1Dict)
obst2Dict = {
    "type": "sphere",
    "geometry": {"position": [2.0, -1.0], "radius": 0.2},
}
sphereObst2 = SphereObstacle(name="simpleSphere", content_dict=obst2Dict)
dynamicObst1Dict = {
    "type": "analyticSphere",
    "geometry": {"trajectory": ["1.1 * t", "-2.0 + 0.1 * t"], "radius": 0.2},
}
dynamicSphereObst1 = DynamicSphereObstacle(
    name="dynamicSphere", content_dict=dynamicObst1Dict
)
splineDict = {
    "degree": 2,
    "controlPoints": [[0.0, 0.0], [5.0, 0.0], [5.0, 5.0]],
    "duration": 10,
}
dynamicObst2Dict = {
    "type": "splineSphere",
    "geometry": {"trajectory": splineDict, "radius": 0.2},
}
dynamicSphereObst2 = DynamicSphereObstacle(
    name="dynamicSphere", content_dict=dynamicObst2Dict
)



from abc import abstractmethod
import numpy as np
from scipy.integrate import odeint
import warnings

from gym import core
from gym.utils import seeding
from gym import spaces


class WrongObservationError(Exception):
    """Exception when observation lays outside the defined observation space.

    This Exception is initiated when an the observation is not within the
    defined observation space. The purpose of this exception is to give
    the user better information about which specific part of the observation
    caused the problem.
    """

    def __init__(self, msg: str, observation: dict, observationSpace):
        """Constructor for error message.

        Parameters
        ----------

        msg: Default error message
        observation: Observation when mismatch occured
        observationSpace: Observation space of environment
        """
        msg_ext = self.get_wrong_observation(observation, observationSpace)
        super().__init__(msg + msg_ext)

    def get_wrong_observation(self, o: dict, os) -> str:
        """Detecting where the error occured.

        Parameters
        ----------

        o: observation
        os: observation space
        """
        msg_ext = ":\n"
        msg_ext += self.check_dict(o, os)
        return msg_ext

    def check_dict(
        self, o_dict: dict, os_dict, depth: int = 1, tabbing: str = ""
    ) -> str:
        """Checking correctness of dictionary observation.

        This methods searches for the cause for wrong observation.
        It loops over all keys in this dictionary and verifies whether
        observation and observation spaces fit together. If this is not
        the case, the concerned key is checked again. As the observation
        might have nested dictionaries, this function is called
        recursively.

        Parameters
        ----------

        o_dict: observation dictionary
        os_dict: observation space dictionary
        depth: current depth of nesting
        tabbing: tabbing for error message
        """
        msg_ext = ""
        for key in o_dict.keys():
            if not os_dict[key].contains(o_dict[key]):
                if isinstance(o_dict[key], dict):
                    msg_ext += tabbing + key + "\n"
                    msg_ext += self.check_dict(
                        o_dict[key],
                        os_dict[key],
                        depth=depth + 1,
                        tabbing=tabbing + "\t",
                    )
                else:
                    msg_ext += self.check_box(
                        o_dict[key], os_dict[key], key, tabbing
                    )
        return msg_ext

    def check_box(
        self, o_box: np.ndarray, os_box, key: str, tabbing: str
    ) -> str:
        """Checks correctness of box observation.

        This methods detects which value in the observation caused the
        error to be raised. Then it updates the error message msg.

        Parameters
        ----------

        o_box: observation box
        os_box: observation space box
        key: key of observation
        tabbing: current tabbing for error message
        """
        msg_ext = tabbing + "Error in " + key + "\n"
        if isinstance(o_box, float):
            val = o_box
            if val < os_box.low[0]:
                msg_ext += f"{tabbing}\t{key}: {val} < {os_box.low[0]}\n"
            elif val > os_box.high[0]:
                msg_ext += f"{tabbing}\t{key}: {val} > {os_box.high[0]}\n"
            return msg_ext

        for i, val in enumerate(o_box):
            if val < os_box.low[i]:
                msg_ext += f"{tabbing}\t{key}[{i}]: {val} < {os_box.low[i]}\n"
            elif val > os_box.high[i]:
                msg_ext += f"{tabbing}\t{key}[{i}]: {val} > {os_box.high[i]}\n"
        return msg_ext


class PlanarEnv(core.Env):
    def __init__(self, render: bool = False, dt=0.01):
        self._viewer = None
        self._state = {
            'joint_state': {
                'position': None,
                'velocity': None
            }
        }
        self._sensor_state = None
        self.seed()
        self._dt = dt
        self._t = 0.0
        self._render = render
        self._obsts = []
        self._goals = []
        self._sensors = []
        self.observation_space = None
        self._n = None

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, n):
        self._n = n

    @n.deleter
    def n(self):
        del self._n

    @abstractmethod
    def set_spaces(self):
        pass

    def dt(self):
        return self._dt

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def add_obstacle(self, obst):
        self._obsts.append(obst)

    def add_goal(self, goal):
        self._goals.append(goal)

    def add_sensor(self, sensor):
        self._sensors.append(sensor)
        observation_space_dict = dict(self.observation_space.spaces)
        observation_space_dict[sensor.name] = sensor.observation_space()
        self.observation_space = spaces.Dict(observation_space_dict)

    def t(self):
        return self._t

    def reset_common(self):
        self._obsts = []
        self._goals = []
        self._sensors = []
        self._t = 0.0

    def reset(self, pos: np.ndarray = None, vel: np.ndarray = None) -> dict:
        self.reset_common()
        if not isinstance(pos, np.ndarray) or not pos.size == self._n:
            pos = np.zeros(self._n)
        if not isinstance(vel, np.ndarray) or not vel.size == self._n:
            vel = np.zeros(self._n)
        self._state['joint_state']['position'] = pos
        self._state['joint_state']['velocity'] = vel
        self._sensor_state = {}
        return self._get_ob()

    def step(self, action: np.ndarray) -> tuple:
        self._action = action
        self.integrate()
        for sensor in self._sensors:
            self._sensor_state = sensor.sense(
                self._state, self._goals, self._obsts, self.t()
            )
        terminal = self._terminal()
        reward = self._reward()
        if self._render:
            self.render()
        return (self._get_ob(), reward, terminal, {})

    @abstractmethod
    def _reward(self):
        pass

    def _get_ob(self):
        observation = dict(self._state)
        observation.update(self._sensor_state)
        if not self.observation_space['joint_state'].contains(observation['joint_state']):
            err = WrongObservationError(
                "The observation does not fit the defined observation space",
                observation['joint_state'],
                self.observation_space['joint_state'],
            )
            warnings.warn(str(err))
        return observation

    @abstractmethod
    def _terminal(self):
        pass

    @abstractmethod
    def continuous_dynamics(self, x, t):
        pass

    def integrate(self):
        self._t += self.dt()
        t = np.arange(0, 2 * self._dt, self._dt)
        x0 = np.concatenate((
                self._state['joint_state']['position'], 
                self._state['joint_state']['velocity'], 
            ))
        ynext = odeint(self.continuous_dynamics, x0, t)
        self._state["joint_state"]["position"] = ynext[1][0 : self._n]
        self._state["joint_state"]["velocity"] = ynext[1][self._n: 2 * self._n]


    @abstractmethod
    def render(self, mode="human"):
        pass

    def render_common(self, bounds):
        from gym.envs.classic_control import (
            rendering,
        )  # pylint: disable=import-outside-toplevel

        if self._state is None:
            return None
        if self._viewer is None:
            if isinstance(bounds, list):
                self._viewer = rendering.Viewer(500, 500)
                self._viewer.set_bounds(
                    -bounds[0], bounds[1], -bounds[1], bounds[1]
                )
            elif isinstance(bounds, dict):
                ratio = (bounds["pos"]["high"][0] - bounds["pos"]["low"][0]) / (
                    bounds["pos"]["high"][1] - bounds["pos"]["low"][1]
                )
                if ratio > 1:
                    window_size = (1000, int(1000 / ratio))
                else:
                    window_size = (int(ratio * 1000), 1000)
                self._viewer = rendering.Viewer(window_size[0], window_size[1])
                self._viewer.set_bounds(
                    bounds["pos"]["low"][0],
                    bounds["pos"]["high"][0],
                    bounds["pos"]["low"][1],
                    bounds["pos"]["high"][1],
                )
        for obst in self._obsts:
            obst.render_gym(self._viewer, rendering, t=self.t())
        for goal in self._goals:
            goal.render_gym(self._viewer, rendering, t=self.t())

    def close(self):
        if self._viewer:
            self._viewer.close()
            self._viewer = None



import numpy as np
from gym import spaces

from planarenvs.ground_robots.envs.ground_robot_env import GroundRobotEnv


class GroundRobotAccEnv(GroundRobotEnv):
    def set_spaces(self):
        self.observation_space = spaces.Dict(
            {"joint_state": spaces.Dict({
                "position": spaces.Box(
                    low=-self._lim_up_pos,
                    high=self._lim_up_pos,
                    dtype=np.float64
                ),
                "velocity": spaces.Box(
                    low=-self._lim_up_vel,
                    high=self._lim_up_vel,
                    dtype=np.float64
                ),
                "forward_velocity": spaces.Box(
                    low=-self._lim_up_rel_vel,
                    high=self._lim_up_rel_vel,
                    dtype=np.float64,
                ),
            })
            }
        )
        self.action_space = spaces.Box(
            low=-self._lim_up_rel_acc,
            high=self._lim_up_rel_acc,
            dtype=np.float64
        )

    def integrate(self):
        super().integrate()
        self._state["xdot"] = self.compute_xdot(
            self._state["joint_state"]["position"], self._state["vel"]
        )

    def continuous_dynamics(self, x, t):
        # state = [x, y, theta, vel_rel, xdot, ydot, thetadot]
        x_pos = x[0:3]
        vel = x[3:5]
        xdot = self.compute_xdot(x_pos, vel)
        veldot = self._action
        xddot = np.zeros(3)
        return np.concatenate((xdot, veldot, xddot))



import numpy as np
import time
from scipy.integrate import odeint
from abc import abstractmethod

from planarenvs.planar_common.planar_env import PlanarEnv


class GroundRobotEnv(PlanarEnv):

    BASE_WIDTH = 1.0  # [m]
    BASE_LENGTH = 1.3  # [m]
    BASE_WHEEL_DIST = 0.6  # [m]
    LINK_MASS_BASE = 5.0  #: [kg] mass of link 1

    MAX_POS_BASE = 5
    MAX_POS_BASE_THETA = 5 * np.pi
    MAX_VEL_BASE = 5
    MAX_VEL_BASE_THETA = 5
    MAX_ACC_BASE = 100
    MAX_ACC_BASE_THETA = 100
    MAX_VEL_FORWARD = 3.0
    MAX_ACC_FORWARD = 100

    def __init__(self, render=False, dt=0.01):
        super().__init__(render=render, dt=dt)
        self._lim_up_pos = np.array(
            [self.MAX_POS_BASE, self.MAX_POS_BASE, self.MAX_POS_BASE_THETA]
        )
        self._lim_up_vel = np.array(
            [self.MAX_VEL_BASE, self.MAX_VEL_BASE, self.MAX_VEL_BASE_THETA]
        )
        self._lim_up_rel_vel = np.array(
            [self.MAX_VEL_FORWARD, self.MAX_VEL_BASE_THETA]
        )
        self._lim_up_acc = np.array(
            [self.MAX_ACC_BASE, self.MAX_ACC_BASE, self.MAX_ACC_BASE_THETA]
        )
        self._lim_up_rel_acc = np.array(
            [self.MAX_ACC_FORWARD, self.MAX_ACC_BASE_THETA]
        )
        self.set_spaces()

    @abstractmethod
    def set_spaces(self):
        pass

    def reset(self, pos=None, vel=None):
        self.reset_common()
        # The velocity is the forward velocity and turning velocity here.
        if not isinstance(pos, np.ndarray) or not pos.size == 3:
            pos = np.zeros(3)
        if not isinstance(vel, np.ndarray) or not vel.size == 2:
            vel = np.zeros(2)
        self._state = {
                "joint_state": {
                    "position": pos,
                    "forward_velocity": vel,
                    "velocity": self.compute_xdot(pos, vel),
                }
        }
        self._sensor_state = {}
        return self._get_ob()

    def compute_xdot(self, x, vel):
        assert x.size == 3
        assert vel.size == 2
        return np.array(
            [
                np.cos(x[2]) * vel[0],
                np.sin(x[2]) * vel[0],
                vel[1],
            ]
        )

    def _reward(self):
        reward = -1.0 if not self._terminal() else 0.0
        return reward

    def integrate(self):
        self._t += self.dt()
        x0 = np.concatenate(
            (
                self._state["joint_state"]["position"],
                self._state["joint_state"]["forward_velocity"],
                self._state["joint_state"]["velocity"])
        )
        t = np.arange(0, 2 * self._dt, self._dt)
        ynext = odeint(self.continuous_dynamics, x0, t)
        self._state["joint_state"]["position"] = ynext[1][0:3]
        self._state["joint_state"]["forward_velocity"] = ynext[1][3:5]
        self._state["joint_state"]["velocity"] = ynext[1][5:8]

    def _terminal(self):
        return False

    @abstractmethod
    def continuous_dynamics(self, x, t):
        pass

    def render(self, mode="human", final=True):
        bound_x = self.MAX_POS_BASE + 1.0
        bound_y = self.MAX_POS_BASE + 1.0
        bounds = [bound_x, bound_y]
        self.render_common(bounds)
        from gym.envs.classic_control import rendering #pylint: disable=import-outside-toplevel

        # drawAxis
        self._viewer.draw_line((-bound_x, 0.0), (bound_x, 0.0))
        self._viewer.draw_line((0.0, -bound_y), (0.0, bound_y))

        p = self._state["joint_state"]["position"][0:2]
        theta = self._state["joint_state"]["position"][2]

        tf = rendering.Transform(rotation=theta, translation=p)

        l, r, t, b = (
            -0.5 * self.BASE_LENGTH,
            0.5 * self.BASE_LENGTH,
            0.5 * self.BASE_WIDTH,
            -0.5 * self.BASE_WIDTH,
        )
        link = self._viewer.draw_polygon([(l, b), (l, t), (r, t), (r, b)])
        yw = self.BASE_WHEEL_DIST / 2.0
        wheelfl = self._viewer.draw_polygon(
            [(0.2, yw), (0.2, yw + 0.1), (0.4, yw + 0.1), (0.4, yw)]
        )
        wheelfr = self._viewer.draw_polygon(
            [(0.2, -yw), (0.2, -yw - 0.1), (0.4, -yw - 0.1), (0.4, -yw)]
        )
        wheelbl = self._viewer.draw_polygon(
            [(-0.2, yw), (-0.2, yw + 0.1), (-0.4, yw + 0.1), (-0.4, yw)]
        )
        wheelbr = self._viewer.draw_polygon(
            [(-0.2, -yw), (-0.2, -yw - 0.1), (-0.4, -yw - 0.1), (-0.4, -yw)]
        )
        wheelfl.add_attr(tf)
        wheelfr.add_attr(tf)
        wheelbl.add_attr(tf)
        wheelbr.add_attr(tf)
        link.set_color(0, 0.8, 0.8)
        link.add_attr(tf)

        if final:
            time.sleep(self.dt())
            return self._viewer.render(return_rgb_array=mode == "rgb_array")



from planarenvs.sensors.sensor_common import Sensor
import numpy as np
from gym import spaces


class FullSensor(Sensor):
    def __init__(self, goal_mask: list, obstacle_mask: list, variance: int= 0.1):
        self._obstacle_mask = obstacle_mask
        self._goal_mask = goal_mask
        self._name = "FullSensor"
        self._noise_variance = variance

    def _reset(self):
        pass

    def observation_size(self) -> tuple:
        return 1, 2

    def observation_space(self):
        return spaces.Box(
            low=-1.0,
            high=1.0,
            shape=self.observation_size(),
            dtype=np.float64,
        )

    def sense(self, state, goals, obstacles: list, t=0):
        sensor_observation = {"goals": [], "obstacles": []}
        for goal in goals:
            exact_observations = list(goal.evaluate_components(self._goal_mask, t).values())
            noisy_observations = []
            for exact_observation in exact_observations:
                if isinstance(exact_observation, np.ndarray):
                    noisy_observation = np.random.normal(exact_observation, self._noise_variance)
                else:
                    noisy_observation = exact_observation
                noisy_observations.append(noisy_observation)
            sensor_observation["goals"].append(noisy_observations)
        for obstacle in obstacles:
            exact_observations = list(obstacle.evaluate_components(self._obstacle_mask, t).values())
            noisy_observations = []
            for exact_observation in exact_observations:
                if isinstance(exact_observation, np.ndarray):
                    noisy_observation = np.random.normal(exact_observation, self._noise_variance)
                else:
                    noisy_observation = exact_observation
                noisy_observations.append(noisy_observation)
            sensor_observation["obstacles"].append(noisy_observations)
        return sensor_observation



import numpy as np
import time
from abc import abstractmethod

from forwardkinematics.planarFks.mobileRobotFk import MobileRobotFk

from planarenvs.planar_common.planar_env import PlanarEnv


class MobileRobotEnv(PlanarEnv):

    BASE_HEIGHT = 1.0  # [m]
    BASE_WIDTH = 1.0  # [m]
    LINK_LENGTH = 1.0  # [m]

    MAX_VEL_BASE = 1
    MAX_POS_BASE = 5.0
    MAX_ACC_BASE = 10.0
    MAX_FOR_BASE = 100.0
    MAX_VEL = 4 * np.pi
    MAX_POS = np.pi
    MAX_ACC = 8 * np.pi
    MAX_TOR = 1000

    def __init__(self, render=False, n=2, dt=0.01):
        super().__init__(render=render, dt=dt)
        self.n = n
        self._lim_up_pos = self.join_limits(
            np.array([self.MAX_POS_BASE]), np.ones(self._n - 1) * self.MAX_POS
        )
        self._lim_up_vel = self.join_limits(
            np.array([self.MAX_VEL_BASE]), np.ones(self._n - 1) * self.MAX_VEL
        )
        self._lim_up_acc = self.join_limits(
            np.array([self.MAX_ACC_BASE]), np.ones(self._n - 1) * self.MAX_ACC
        )
        self._lim_up_tor = self.join_limits(
            np.array([self.MAX_FOR_BASE]), np.ones(self._n - 1) * self.MAX_TOR
        )
        self.set_spaces()
        self._fk = MobileRobotFk(self._n, baseHeight=self.BASE_HEIGHT)

    def join_limits(self, lim_bas, lim_arm):
        return np.concatenate((lim_bas, lim_arm))

    @abstractmethod
    def set_spaces(self):
        pass

    def _get_ob(self):
        return self._state

    def _terminal(self):
        if (
            self._state["joint_state"]["position"][0] > self.MAX_POS_BASE
            or self._state["joint_state"]["position"][0] < -self.MAX_POS_BASE
        ):
            return True
        return False

    def _reward(self):
        reward = -1.0 if not self._terminal() else 0.0
        return reward

    @abstractmethod
    def continuous_dynamics(self, x, t):
        pass

    def render(self, mode="human"):
        bound = self.MAX_POS_BASE + 1.0
        bounds = [bound, bound]
        self.render_common(bounds)

        # drawAxis
        self._viewer.draw_line((-5.5, 0), (5.5, 0))

        self.render_base()
        for i in range(1, self._n):
            self.render_link(i)
        self.render_end_effector()
        time.sleep(self.dt())
        return self._viewer.render(return_rgb_array=mode == "rgb_array")

    def render_base(self):
        from gym.envs.classic_control import rendering #pylint: disable=import-outside-toplevel

        l, r, t, b = (
            -0.5 * self.BASE_WIDTH,
            0.5 * self.BASE_WIDTH,
            0.5 * self.BASE_HEIGHT,
            -0.5 * self.BASE_HEIGHT,
        )
        tf = rendering.Transform(
            rotation=0,
            translation=(self._state["joint_state"]["position"][0], 0.5 * self.BASE_HEIGHT),
        )
        link = self._viewer.draw_polygon([(l, b), (l, t), (r, t), (r, b)])
        link.set_color(0, 0.8, 0.8)
        link.add_attr(tf)
        base = self._viewer.draw_polygon(
            [(-0.2, -0.2), (0.0, 0.0), (0.0, 0.0), (0.2, -0.2)]
        )
        base_joint = self._viewer.draw_circle(0.10)
        base_joint.set_color(0.8, 0.8, 0)
        tf0 = rendering.Transform(
            rotation=0,
            translation=(self._state["joint_state"]["position"][0], self.BASE_HEIGHT + 0.2),
        )
        base_joint.add_attr(tf0)
        base.add_attr(tf0)

    def render_link(self, i):
        from gym.envs.classic_control import rendering #pylint: disable=import-outside-toplevel

        l, r, t, b = 0, self.LINK_LENGTH, 0.01, -0.01
        fk = self._fk.fk(self._state["joint_state"]["position"], i)
        tf = rendering.Transform(rotation=fk[2], translation=fk[0:2])
        link = self._viewer.draw_polygon([(l, b), (l, t), (r, t), (r, b)])
        link.set_color(0, 0.8, 0.8)
        joint = self._viewer.draw_circle(0.10)
        joint.set_color(0.8, 0.8, 0)
        link.add_attr(tf)
        joint.add_attr(tf)

    def render_end_effector(self):
        from gym.envs.classic_control import rendering #pylint: disable=import-outside-toplevel

        fk = self._fk.fk(self._state["joint_state"]["position"], self._n)
        tf = rendering.Transform(rotation=fk[2], translation=fk[0:2])
        eejoint = self._viewer.draw_circle(0.10)
        eejoint.set_color(0.8, 0.8, 0)
        eejoint.add_attr(tf)



import numpy as np
from gym import spaces

from planarenvs.ground_robots.envs.ground_robot_arm_env import GroundRobotArmEnv


class GroundRobotArmAccEnv(GroundRobotArmEnv):
    def set_spaces(self):
        self.observation_space = spaces.Dict(
                {"joint_state": spaces.Dict({
                    "position": spaces.Box(
                        low=np.concatenate((-self._lim_up_pos, -self._lim_up_arm_pos)),
                        high=np.concatenate((self._lim_up_pos, self._lim_up_arm_pos)),
                        dtype=np.float64
                ),
                    "velocity": spaces.Box(
                        low=np.concatenate((-self._lim_up_vel, -self._lim_up_arm_vel)),
                        high=np.concatenate((self._lim_up_vel, self._lim_up_arm_vel)),
                        dtype=np.float64
                ),
                "forward_velocity": spaces.Box(
                    low=-self._lim_up_rel_vel,
                    high=self._lim_up_rel_vel,
                    dtype=np.float64,
                ),
            })
            }
        )
        a = np.concatenate((self._lim_up_rel_acc, self._lim_up_arm_acc))
        self.action_space = spaces.Box(low=-a, high=a, dtype=np.float64)

    def continuous_dynamics(self, x, t):
        # state = [x, y, theta, vel, xdot_base, q, qdot)
        x_pos = x[0:3]
        vel = x[3:5]
        xdot_base = self.compute_xdot(x_pos, vel)
        veldot = self._action[0:2]
        qddot = self._action[2:]
        qdot = x[8 + self._n_arm :]
        xddot = np.zeros(3)
        return np.concatenate((xdot_base, veldot, xddot, qdot, qddot))



from gym.envs.registration import register

register(
    id="point-robot-acc-v0",
    entry_point="planarenvs.point_robot.envs:PointRobotAccEnv",
)
register(
    id="point-robot-vel-v0",
    entry_point="planarenvs.point_robot.envs:PointRobotVelEnv",
)



# pylint: disable-all
from planarenvs.n_link_reacher.resources.createDynamics import createDynamics


class AccController(object):

    """Simple forward controller acceleration to torque"""

    def __init__(self, n, l, m, g, k):
        _, _, self.tau_fun = createDynamics(n)
        self._n = n
        self._l = l
        self._m = m
        self._g = g
        self._k = k

    def control(self, q, qdot, qddot):
        tau = self.tau_fun(q, qdot, self._l, self._m, self._g, self._k, qddot)
        return tau



import numpy as np
import time
from scipy.integrate import odeint

from planarenvs.ground_robots.envs.ground_robot_env import GroundRobotEnv


class GroundRobotArmEnv(GroundRobotEnv):
    LINK_LENGTH = 1.0
    MAX_ARM_VEL = 4 * np.pi
    MAX_ARM_POS = 5 * np.pi
    MAX_ARM_ACC = 9 * np.pi

    def __init__(self, render=False, dt=0.01, n_arm=1):
        self._n_arm = n_arm
        self._lim_up_arm_pos = np.ones(self._n_arm) * self.MAX_ARM_POS
        self._lim_up_arm_vel = np.ones(self._n_arm) * self.MAX_ARM_VEL
        self._lim_up_arm_acc = np.ones(self._n_arm) * self.MAX_ARM_ACC
        super().__init__(render=render, dt=dt)

    def integrate(self):
        self._t += self.dt()
        x0 = np.concatenate(
            (
                self._state["joint_state"]["position"],
                self._state["joint_state"]["forward_velocity"],
                self._state["joint_state"]["velocity"],
            )
        )
        t = np.arange(0, 2 * self._dt, self._dt)
        ynext = odeint(self.continuous_dynamics, x0, t)
        self._state["joint_state"]["position"] = np.concatenate((ynext[1][0:3], ynext[1][8: 8 + self._n_arm]))
        self._state["joint_state"]["forward_velocity"] = ynext[1][3:5]
        self._state["joint_state"]["velocity"] = np.concatenate((ynext[1][5:8], ynext[1][8+self._n_arm:]))

    def reset(self, pos=None, vel=None):
        self.reset_common()
        # The velocity is the forward velocity and turning velocity here.
        if not isinstance(pos, np.ndarray) or not pos.size == 3 + self._n_arm:
            pos = np.zeros(3 + self._n_arm)
        if not isinstance(vel, np.ndarray) or not vel.size == 2 + self._n_arm:
            vel = np.zeros(2 + self._n_arm)
        xdot_base = self.compute_xdot(pos[0:3], vel[0:2])
        self._state = {
                "joint_state": {
                    "position": pos,
                    "forward_velocity": vel[0:2],
                    "velocity": np.concatenate((xdot_base, vel[2:])),
                }
            }
        self._sensor_state = {}
        return self._get_ob()

    def render(self, mode="human"):
        from gym.envs.classic_control import (  # pylint: disable=import-outside-toplevel
            rendering,
        )

        super().render(mode=mode, final=False)

        # arm
        p = self._state["joint_state"]["position"][0:2]
        theta = self._state["joint_state"]["position"][2]
        q = self._state["joint_state"]["position"][3:]
        l, r, t, b = 0, self.LINK_LENGTH, 0.05, -0.05
        p_arm = p + 0.2 * np.array([np.cos(theta), np.sin(theta)])
        tf_arm = rendering.Transform(rotation=theta + q, translation=p_arm)
        link = self._viewer.draw_polygon([(l, b), (l, t), (r, t), (r, b)])
        link.set_color(0, 0.2, 0.8)
        link.add_attr(tf_arm)
        time.sleep(self.dt())

        return self._viewer.render(return_rgb_array=mode == "rgb_array")



from abc import ABC, abstractmethod
import numpy as np
from gym import spaces


def dist2circ(
    point_pos: np.ndarray, circle_pos: np.ndarray, circle_radius: float
) -> np.ndarray:
    diff = point_pos - circle_pos
    dist_to_center = np.linalg.norm(diff)
    if dist_to_center < circle_radius:
        return np.zeros(2)
    diff = diff * (dist_to_center - circle_radius) / dist_to_center
    return diff


class Sensor(ABC):
    def __init__(self, nb_observations, lim_sensor=10):
        self._nb_observations = nb_observations
        self._lim_sensor = lim_sensor
        self._observation = []
        self._name = "Sensor"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @name.deleter
    def name(self):
        del self._name

    def observation_size(self) -> tuple:
        return self._nb_observations, 2

    def observation_space(self):
        return spaces.Box(
            low=-self._lim_sensor,
            high=self._lim_sensor,
            shape=self.observation_size(),
            dtype=np.float64,
        )

    @abstractmethod
    def _reset(self):
        pass

    @abstractmethod
    def sense(self, s, goals, obstacles, t=0):
        pass



from gym.envs.registration import register

register(
    id="mobile-robot-vel-v0",
    entry_point="planarenvs.mobile_robot.envs:MobileRobotVelEnv",
)
register(
    id="mobile-robot-acc-v0",
    entry_point="planarenvs.mobile_robot.envs:MobileRobotAccEnv",
)



from planarenvs.sensors.sensor_common import Sensor
from planarenvs.sensors.sensor_common import dist2circ
import numpy as np
import operator


class ObstacleSensor(Sensor):
    def __init__(self, nb_obstacles=0, lim_sensor=10, mode="position"):
        super().__init__(nb_observations=nb_obstacles, lim_sensor=lim_sensor)
        self._observation = (
            np.ones([self._nb_observations, 2]) * self._lim_sensor
        )
        self._mode = mode
        self.name = mode

    @Sensor.name.setter
    def name(self, mode):
        if mode == "position":
            self._name = "ObstaclePosition"
        elif mode == "distance":
            self._name = "ObstacleDistance"

    def _reset(self):
        self._observation[:] = self._lim_sensor

    def sense(
        self, state: dict, goals, obstacles, t=0
    ) -> dict:  # pylint: disable=unused-argument
        self._reset()
        if self._mode == "position":
            for idx, obst in enumerate(obstacles):
                if idx >= self._nb_observations:
                    break
                self._observation[idx] = obst.position(t=t)
        elif self._mode == "distance":
            self._reset()
            current_position = state["joint_state"]["position"]  # only true for point_robot
            obstacles_info = []
            for id_obstacle, obstacle in enumerate(obstacles):
                obstacle_position = obstacle.position(t=t)
                obstacle_distance = dist2circ(
                    current_position,
                    obstacle_position,
                    obstacle.radius(),
                )
                absolute_distance = abs(obstacle_distance[0]) + abs(
                    obstacle_distance[1]
                )
                obstacles_info.append(
                    {
                        "id": id_obstacle,
                        "distance": obstacle_distance,
                        "position": obstacle_position,
                        "absolute_distance": absolute_distance,
                    }
                )
            obstacles_info.sort(
                reverse=False, key=operator.itemgetter("absolute_distance")
            )
            for idx, obst in enumerate(obstacles_info):
                if idx >= self._nb_observations:
                    break
                self._observation[idx] = obst["distance"]

        return self._observation.clip(-self._lim_sensor, self._lim_sensor)
